/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#define _REENTRANT
#undef _GNU_SOURCE
#define _GNU_SOURCE

#include "../config.h"
#include "Connection.h"
#include "filters.h"
#include <DSHashTable.h>
#include <DSConfigFile.h>
#include <DSMutex.h>
#include <time.h>
#include <sys/time.h>
#include <unistd.h>
#include <netdb.h>
#include "Shared.h"
#include "Locks.h"
#include "moddecl.h"
#include "loggers.h"

extern DSHashTable versions;
extern DSHashTable resolvers;
extern DSHashTable revresolve;
extern DSList filters;
extern DSList loggers;

Connection::Connection(DSNetwork * thisnet)
{
	SOCKET s;
	PI_SA sa;
	sl_t sl;
	net=thisnet;
	user=NULL;

	buffer=NULL;
	bufflen=0;

	s=net->getSocket();
	sl=sizeof(sa);
	if (getpeername(s, (SOCKADDR *)&sa, &sl)==0) {
		source.setAddress((SOCKADDR *)&sa);
	}
	if (getsockname(s, (SOCKADDR *)&sa, &sl)==0) {
		source_on_server.setAddress((SOCKADDR *)&sa);
	}

	nExternalSendBytes=0;
	nExternalRecvBytes=0;
}

Connection::Connection(SOCKADDR * sa,char * udppacket, int udplen)
{
	net=NULL;
	user=NULL;

	this->buffer=udppacket;
	this->bufflen=udplen;
	buff_upto=0;

	source.setAddress(sa);
}

Connection::~Connection()
{
	delete net;
}

unsigned long Connection::getDownloadBytes()
{
	return nExternalRecvBytes;
}

unsigned long Connection::getUploadBytes()
{
	return nExternalSendBytes;
}

char * Connection::getRemainingData()
{
	if (!buffer) return NULL;
	return &buffer[buff_upto];
}

int Connection::getRemainingLength()
{
	if (!buffer) return 0;
	return bufflen-buff_upto;
}

char * Connection::getCompleteData()
{
	if (!buffer) return NULL;
	return buffer;
}

int Connection::getCompleteLength()
{
	if (!buffer) return 0;
	return buff_upto;
}

char * Connection::getUser()
{
	return user;
}

void Connection::setUser(char * newuser)
{
	if (user) free(user);
	if (newuser==NULL) {
		user=NULL;
	} else {
		user=(char *)malloc(strlen(newuser)+1);
		strcpy(user,newuser);
	}
}

int Connection::getSourceAddressType()
{
	return source.getAddressType();
}

unsigned int Connection::getSourcePort()
{
	return source.getPort();
}

int Connection::getSourceAddressLength()
{
	return source.getAddressLength();
}

char * Connection::getSourceAddress()
{
	return source.getAddress();
}

char * Connection::getSourceString()
{
	return source.getString();
}

char * Connection::getSourceAddrString()
{
	return source.getAddressString();
}

BOOL Connection::getSourceSockaddr(SOCKADDR ** sasrc, sl_t * len)
{
	return source.getSockaddr(sasrc,len);
}

int Connection::getSrvSourceAddressType()
{
	return source_on_server.getAddressType();
}

unsigned int Connection::getSrvSourcePort()
{
	return source_on_server.getPort();
}

int Connection::getSrvSourceAddressLength()
{
	return source_on_server.getAddressLength();
}

char * Connection::getSrvSourceAddress()
{
	return source_on_server.getAddress();
}

char * Connection::getSrvSourceString()
{
	return source_on_server.getString();
}

char * Connection::getSrvSourceAddrString()
{
	return source_on_server.getAddressString();
}

BOOL Connection::getSrvSourceSockaddr(SOCKADDR ** sasrc, sl_t * len)
{
	return source_on_server.getSockaddr(sasrc,len);
}

int Connection::getDestAddressType()
{
	return dest.getAddressType();
}

unsigned int Connection::getDestPort()
{
	return dest.getPort();
}

int Connection::getDestAddressLength()
{
	return dest.getAddressLength();
}

char * Connection::getDestAddress()
{
	return dest.getAddress();
}

BOOL Connection::getDestNullAddr()
{
	return dest.getNullAddr();
}

char * Connection::getDestString()
{
	return dest.getString();
}

char * Connection::getDestAddrString()
{
	return dest.getAddressString();
}

BOOL Connection::getDestSockaddr(SOCKADDR ** sadest, sl_t * len)
{
	return dest.getSockaddr(sadest,len);
}

void Connection::setDestAddress(char * addr, int len)
{
	dest.setAddress(addr,len);
}

BOOL Connection::setDestHostname(char * host)
{
	HOSTENT * phe;
#if defined(HAVE_GETHOSTBYNAME_R)||defined(HAVE_NSL_GETHOSTBYNAME_R)
	HOSTENT realhe;
	char buf[1024];
	int perrno;
#ifdef BROKEN_GETHOSTBYNAME
	phe=gethostbyname_r(host, &realhe, buf, sizeof(buf), &perrno);
#else
	gethostbyname_r(host, &realhe, buf, sizeof(buf), &phe, &perrno);
#endif
#else
	gethostbyname_lock.lock();
	phe=gethostbyname(host);
#endif
	if (!phe) {
#ifdef WITH_DEBUG
		printf("Failed to resolve hostname\n");
#endif
#if (!defined(HAVE_GETHOSTBYNAME_R))&&(!defined(HAVE_NSL_GETHOSTBYNAME_R))
		gethostbyname_lock.unlock();
#endif
		return FALSE;
	}
	setDestAddressType(phe->h_addrtype);
	setDestAddress(phe->h_addr_list[0],phe->h_length);
#if (!defined(HAVE_GETHOSTBYNAME_R))&&(!defined(HAVE_NSL_GETHOSTBYNAME_R))
	gethostbyname_lock.unlock();
#endif
	return TRUE;
}


void Connection::setDestPort(unsigned int port)
{
	dest.setPort(port);
}

void Connection::setDestAddressType(int newtype)
{
	dest.setAddressType(newtype);
}


BOOL Connection::AllowConnection()
{
	SOCKET s;
	PI_SA sa;
	sl_t sl;
	s=net->getSocket();
	sl=sizeof(sa);
	if (getsockname(s, (SOCKADDR *)&sa, &sl)==0) {
		return IsAllowed(TYPE_SOURCE);
	} else {
		return FALSE;
	}
}


BOOL Connection::IsAllowed(int type)
{
	DSListElement * le;
	proc_filter filter;
	BOOL ret;
	for (le=filters.getHead();le!=NULL;le=le->getNext()) {
		filter=(proc_filter)le->getDataPtr();
		ret=filter(this,type);
		if (ret==FALSE) {
#ifdef WITH_DEBUG
			printf("Connection dropped due to filtering\n");
#endif
			return FALSE;
		}
	}
	return TRUE;
}

BOOL Connection::sendData(const char * szData, int nSize)
{
	if (buffer) {
		int wanttosend;
		wanttosend=nSize;
		if (buff_upto+wanttosend>=bufflen) {
			wanttosend=bufflen-buff_upto;
		}
		memcpy(&buffer[buff_upto],szData,wanttosend);
		buff_upto+=wanttosend;
		if (nSize==wanttosend) return TRUE;
		return FALSE;
	} else {
		return net->sendData(szData,nSize);
	}
}

BOOL Connection::getChar(unsigned char * ch)
{
	if (buffer) {
		if (buff_upto<bufflen) {
			*ch=buffer[buff_upto];
			buff_upto++;
			return TRUE;
		}
		return FALSE;
	} else
		return net->getChar(ch);
}

BOOL Connection::getChar(char * ch)
{
	return getChar((unsigned char *)ch);
}

BOOL Connection::getSlab(char * szBuffer, int nSize)
{
	if (buffer) {
		int wanttoread;
		wanttoread=nSize;
		if ((buff_upto+wanttoread)>=bufflen) {
			wanttoread=bufflen-buff_upto;
		}
		memcpy(szBuffer,&buffer[buff_upto],wanttoread);
		buff_upto+=wanttoread;
		if (nSize==wanttoread) return TRUE;
		return FALSE;
	} else
		return net->getSlab(szBuffer, nSize);
}


BOOL Connection::sendResponsev4(char rep, unsigned long addr,
		unsigned short port)
{
	char szResp[8];
	szResp[0]=0x00;
	szResp[1]=rep;
	memcpy(&szResp[2],&port,2);
	memcpy(&szResp[4],&addr,4);
	sendData(szResp,8);
	return TRUE;
}

BOOL Connection::sendResponsev5(char rep, SOCKADDR * sa)
{
	char szResp[100];
	char szTemp[10];
	char * oldbuffer;
	int oldbuflen;
	int oldbuff_upto;
	int havetosend;
	proc_revresolve revres;

	/* Take a copy of the old buffer info */
	oldbuffer=buffer;
	oldbuflen=bufflen;
	oldbuff_upto=buff_upto;

	/* Set up writes to temp location */
	buffer=szResp;
	bufflen=sizeof(szResp);
	buff_upto=0;

	revres=NULL;
	szTemp[0]=0x05;
	szTemp[1]=rep;
	szTemp[2]=0x00;
	sendData(szTemp,3);
	if (sa)
		revres=(proc_revresolve)revresolve.getPtrValue(sa->sa_family);

	if (revres==NULL) {
		int i;
		szTemp[0]=0x01;
		for (i=1;i<7;i++) szTemp[i]=0x00;
		sendData(szTemp,7);
	} else {
		revres(this, sa);
	}

	havetosend=buff_upto;

	/* Write to wherever the old buffer was, hopefully network */
	buffer=oldbuffer;
	bufflen=oldbuflen;
	buff_upto=oldbuff_upto;

	sendData(szResp, havetosend);
	return TRUE;
}

SOCKET Connection::getSocket()
{
	return net->getSocket();
}

void * Connection::getResolver(int nResolver)
{
	return resolvers.getPtrValue(nResolver);
}

void * Connection::getRevResolver(int nResolver)
{
	return revresolve.getPtrValue(nResolver);
}

BOOL Connection::forwardData(SOCKET two)
{
	SOCKET one;
	BOOL finished;
	char szBuffer[1024];
	SOCKET biggest;
	int len;
	int throttle;
	time_t start;
	fd_set fds;

	one=net->getSocket();
	throttle=config.getValueInt("boot","throttle",0);
	time(&start);
	
	finished = FALSE;
	while (!finished)
	{
		if (throttle>0) {
			int currentrate;
			time_t now;
			time(&now);
			if (now==start) {
				currentrate=throttle;
			} else {
				currentrate=(nExternalSendBytes+nExternalRecvBytes)/(now-start);
			}
			if (currentrate>=throttle) {
				int sleepfor;
				sleepfor=currentrate/throttle+1;
				sleep(sleepfor);
			}
		}
		FD_ZERO (&fds);
		FD_SET (one, &fds);
		FD_SET (two, &fds);
		if (one > two)
			biggest = one;

		else
			biggest = two;
		biggest++;
		/* Wait for message from either side */
		select (biggest, &fds, NULL, NULL, NULL);
		if (FD_ISSET (one, &fds))
		{
			/* A message from the LAN, forward to the server */
			len = recv (one, szBuffer,
				    sizeof(szBuffer) - 1, 0);
			if (len <= 0)
			{
#ifdef WITH_DEBUG
				printf("Connection closed by local\n");
#endif
				/* Connection closed */
				finished = TRUE;
			}
			else
			{
				nExternalSendBytes+=len;
				szBuffer[len] = '\0';
				send (two, szBuffer, len, 0);
#ifdef WITH_DEBUG
				printf("Send %i bytes local->remote\n",len);
#endif
			}
		}
		if (FD_ISSET (two, &fds))
		{
			/* A message from the server, forward to the LAN */
			len = recv (two, szBuffer,
				    sizeof(szBuffer) - 1, 0);
			if (len <= 0)
			{
#ifdef WITH_DEBUG
				printf("Connection closed by remote\n");
#endif
				/* Connection closed */
				finished = TRUE;
			}
			else
			{
				nExternalRecvBytes+=len;
				szBuffer[len] = '\0';
				send (one, szBuffer, len, 0);
#ifdef WITH_DEBUG
				printf("Send %i bytes remote->local\n",len);
#endif
			}
		}
	}
	return TRUE;
}



BOOL Connection::log(int type)
{
	DSListElement * le;
	proc_logger logger;
	BOOL ret;
	for (le=loggers.getHead();le!=NULL;le=le->getNext()) {
		if (le->getKeyInt()==(unsigned int)type) {
			logger=(proc_logger)le->getDataPtr();
			ret=logger(this,LOG_EVT_LOG);
			if (ret==FALSE) {
#ifdef WITH_DEBUG
				printf("An attempt to log a condition failed\n");
#endif
				return FALSE;
			}
		}
	}
	return TRUE;
}




BOOL Connection::ChildThread()
{
	unsigned char version;
	proc_version processor;
	processor=NULL;
	if (!AllowConnection()) goto barfed;
	if (!net->getChar((char *)&version)) goto barfed;
#ifdef WITH_DEBUG
	printf("Version: %x\n",version);
#endif
	processor=(proc_version)versions.getPtrValue(version);
	if (processor!=NULL) 
		processor(this);

barfed:
	log(LOG_TYPE_CONNECTIONCLOSE);

#ifdef WITH_DEBUG
	printf("Cleaning up from connection\n");
#endif
	net->terminate();
	return FALSE;
}

void * ChildThread(void * conn)
{
	Connection * connection;
	connection=(Connection *)conn;
	connection->ChildThread();
	delete connection;
	return NULL;
}


