/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#include "../../../config.h"
#include <DSNetwork.h>
#include <DSConfigFile.h>
#include <DSBinaryTree.h>
#ifdef HAVE_NETDB_H
#include <netdb.h>
#endif
#include "../../Shared.h"
#include "../../filters.h"
#include "../../Connection.h"

DSBinaryTree *sourceipfilters;
DSBinaryTree *destipfilters;

typedef struct stBlock {
	int whattodo;
	DSBinaryTree * tree;
} stBlock;


void hostblocker(char * entry, char * value, void * extra)
{
	HOSTENT * phe;
	char * bits;
	unsigned int host;
	int netbits;
	stBlock * data;
	data = (stBlock *)extra;
	/* No locking necessary as function only called in initialisation */
	bits=strchr(entry, '/');
	if (bits) {
		netbits=atoi(bits+1);
		if (netbits>32) netbits=32;
		bits[0]='\0';
	} else {
		netbits=32;
	}
	phe=gethostbyname(entry);
	if (phe) {
		/* FIXME: IPv4ism */
		host=ntohl(*(int *)phe->h_addr);
#ifdef WITH_DEBUG
		printf("Adding %x/%i to tree\n",host,netbits);
#endif
		data->tree->insert(host, netbits, data->whattodo);
	} else {
#ifdef WITH_DEBUG
		printf("Could not add filter entry %s!\n",entry);
#endif
	}
	if (bits)
		bits[0]='/';
}


/*
 * Load in a config file and perform the necessary conversions on users/groups
 * to allow it to be used to filter.
 */
void load_filter_config()
{
	stBlock blk;
	sourceipfilters=new DSBinaryTree();
	destipfilters=new DSBinaryTree();
	blk.tree=sourceipfilters;
	blk.whattodo=FILTER_ALLOW;
	config.forEach("allow source IPv4",hostblocker,&blk);
	blk.whattodo=FILTER_DENY;
	config.forEach("block source IPv4",hostblocker,&blk);

	blk.tree=destipfilters;
	blk.whattodo=FILTER_ALLOW;
	config.forEach("allow destination IPv4",hostblocker,&blk);
	blk.whattodo=FILTER_DENY;
	config.forEach("block destination IPv4",hostblocker,&blk);
}

BOOL cpp_filter(Connection * conn,int type)
{
	unsigned long host;
	int ret;
	ret=FILTER_DEFAULT;

	if (type==TYPE_SOURCE) {
		host=ntohl(*(unsigned long *)conn->getSourceAddress());
		ret=sourceipfilters->getClosestInt(host,32);
	}
	if (type==TYPE_DEST) {
		host=ntohl(*(unsigned long *)conn->getDestAddress());
		ret=destipfilters->getClosestInt(host,32);
	}
#ifdef WITH_DEBUG
	printf("Filter reports %i to type %i\n",ret,type);
#endif
	if (ret==FILTER_DEFAULT) {
		if (type==TYPE_SOURCE) {
			return config.getValueInt("defaults","allow unknown source IPv4",0);
		}
		if (type==TYPE_DEST) {
			return config.getValueInt("defaults","allow unknown destination IPv4",0);
		}
		/* If unknown type of filter, we don't support, so allow */
		return TRUE;
	}
	if (ret==FILTER_ALLOW) return TRUE;
	return FALSE;
}

extern "C" {
int libipv4_address_type()
{
	load_filter_config();
	return AF_INET;
}

BOOL libipv4_filter(void * conn,int type)
{
	return cpp_filter((Connection *)conn, type);
}
}
