/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#define _REENTRANT
#include "../../config.h"
#ifdef HAVE_TIME_H
#include <time.h>
#endif
#include <DSNetwork.h>
#include <DSMutex.h>
#include "../Connection.h"
#include "../loggers.h"
#include "../Locks.h"

FILE * logfile;

DSMutex connection_filelock;

BOOL connection_real_log(Connection * conn)
{
	char * strSourceAddr;
	char * strDestAddr;
	char * strUser;
	time_t now;
	char strTime[100];
#ifdef HAVE_LOCALTIME_R
	struct tm realtm;
#endif
	struct tm * tm;
	time(&now);

#ifdef HAVE_LOCALTIME_R
	tm=localtime_r(&now, &realtm);
#else
	localtime_lock.lock();
	tm=localtime(&now);
#endif
	strftime(strTime, sizeof(strTime), "%a, %d %b %Y %H:%M:%S", tm);
#ifndef HAVE_LOCALTIME_R
	localtime_lock.unlock();
#endif

	connection_filelock.lock();
	if (logfile) {
		strSourceAddr=conn->getSourceString();
		strDestAddr=conn->getDestString();
		strUser=conn->getUser();

		if (strUser) {
			fprintf(logfile, "%s %s %s %s\n", strTime,
				strSourceAddr, strDestAddr, strUser);
		} else {
			fprintf(logfile, "%s %s %s\n", strTime, 
				strSourceAddr, strDestAddr);
		}
		connection_filelock.unlock();

		free(strSourceAddr);
		free(strDestAddr);
		return TRUE;
	} else {
		connection_filelock.unlock();
		return FALSE;
	}
}

BOOL connection_log(Connection * conn,int event)
{
	switch(event) {
		case LOG_EVT_SERVERSTART:
		connection_filelock.lock();
		logfile=fopen(logdir "/connection.log","a");
		connection_filelock.unlock();
		break;
		case LOG_EVT_SERVERCLOSE:
		connection_filelock.lock();
		if (logfile) fclose(logfile);
		connection_filelock.unlock();
		break;
		case LOG_EVT_LOG:
		return connection_real_log(conn);
		break;
	}
	return TRUE;
}

extern "C" {
int libconn_logger_type()
{
	return LOG_TYPE_CONNECTIONESTABLISHED;
}

BOOL libconn_log(void * conn,int event)
{
	return connection_log((Connection *)conn,event);
}
}
