/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#include "../../../config.h"
#include <DSNetwork.h>
#include <DSConfigFile.h>
#ifdef HAVE_NETINET_IN_H
#include <netinet/in.h>
#endif
#ifdef HAVE_SYS_SELECT_H
#include <sys/select.h>
#endif
#ifdef HAVE_UNISTD_H
#include <unistd.h>
#endif
#include "../../Shared.h"
#include "../../Connection.h"
#include "../../loggers.h"

extern "C" {
int libv5bind_v5command_type()
{
	return 0x02;
}
}

/* This function reports to the client a socket's specs. */
void notifyclient(Connection * conn, SOCKET s,SOCKADDR * sout, sl_t * soutlen)
{
	if (getsockname(s, (SOCKADDR *)sout, soutlen)==0) {
#ifdef WITH_DEBUG
		printf("Bind succeeded, interface known\n");
#endif
		conn->sendResponsev5(0x00, sout);
	} else {
#ifdef WITH_DEBUG
		printf("Bind succeeded, interface unknown\n");
#endif
		conn->sendResponsev5(0x00);
	}
}

BOOL ver5bind_command(Connection * conn)
{
	SOCKET remotelisten, remotetalk;
	PI_SA sout;
	sl_t soutlen;
	SOCKADDR * addr;
	sl_t addrlen;

	if (!conn->getDestSockaddr(&addr, &addrlen)) return FALSE;
#ifdef WITH_DEBUG
	printf("Socket family: %x\n",addr->sa_family);
#endif
	remotelisten=socket(addr->sa_family, SOCK_STREAM, 0);
#ifdef WITH_DEBUG
	{
		char * dest;
		dest=conn->getDestString();
		printf("Binding to %s\n",dest);
		free(dest);
	}
#endif
	/* FIXME: This hack listens on all interfaces on any port.
	 * Is this really what the spec had in mind? */
	memset(&sout, 0, sizeof(sout));
	((SOCKADDR *)&sout)->sa_family=addr->sa_family;
	soutlen=addrlen;
	if ((bind(remotelisten, (SOCKADDR *)&sout, soutlen)!=0)||
		(listen(remotelisten, 2)!=0)) {
#ifdef WITH_DEBUG
		printf("Bind failed\n");
		perror("bind");
#endif
		CloseAndUnlink(remotelisten);
		conn->sendResponsev5(0x01);
		free(addr);
		return FALSE;
	}
	free(addr);
	soutlen=sizeof(sout);
	notifyclient(conn, remotelisten, (SOCKADDR *)&sout, &soutlen);

	/* Now to wait for incoming connection */
	soutlen=sizeof(sout);
	remotetalk=accept(remotelisten, (SOCKADDR *)&sout, &soutlen);
	CloseAndUnlink(remotelisten);
	if (remotetalk<0) {
		/* Whoa, accept() failed.  Better tell client */
		conn->sendResponsev5(0x01);
		return FALSE;
	}
	AddressInfo ai((SOCKADDR *)&sout);
	int correct;
	char * whoArrived;
	char * whoShould;
	whoArrived=ai.getAddressString();
	whoShould=conn->getDestAddrString();
	correct=strcmp(whoArrived,whoShould);
	free(whoArrived);
	free(whoShould);
	if (correct!=0) {
		/* Somebody connected who shouldn't have. */
		conn->sendResponsev5(0x01);
		return FALSE;
	}

	/* Tell client who just arrived */
	soutlen=sizeof(sout);
	notifyclient(conn, remotetalk, (SOCKADDR *)&sout, &soutlen);

	conn->log(LOG_TYPE_CONNECTIONESTABLISHED);

	/* As with connect, start forwarding data. */
	conn->forwardData(remotetalk);
	closesocket(remotetalk);

	return TRUE;
}

extern "C" {
BOOL libv5bind_v5command(void * conn)
{
	return ver5bind_command((Connection *)conn);
}
}


