/* ANTINAT
 * =======
 * This software is Copyright (c) 2003 Malcolm Smith.
 * No warranty is provided, including but not limited to
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * This code is licenced subject to the GNU General
 * Public Licence (GPL).  See the COPYING file for more.
 */

#include "../../config.h"
#include <sys/types.h>
#ifdef HAVE_CTYPE_H
#include <ctype.h>
#endif
#ifdef HAVE_LIMITS_H
#include <limits.h>
#endif
#include <DSNetwork.h>
#include <DSConfigFile.h>
#include <DSHashTable.h>
#include <DSComponent.h>
#include "../Shared.h"
#include "../Connection.h"
#include "../filters.h"
#include "../moddecl.h"

/* These structures should not be written to from within a connection thread,
 * only the master thread.  So they're not really protected. */
DSHashTable *authenticators;
DSHashTable *v5_commands;

BOOL Initialize(Connection * conn)
{
	unsigned char nMethods;
	unsigned char method;
	int i;
	proc_auth processor;
	void * tmp;
	unsigned char chosenmethod;
	char szResp[2];
	processor=NULL;
	chosenmethod=0xff;

	if (!conn->getChar(&nMethods)) return FALSE;
#ifdef WITH_DEBUG
	printf("Number of Methods: %i\n",nMethods);
#endif
	for (i=0;i<nMethods;i++) {
		if (!conn->getChar(&method)) return FALSE;
#ifdef WITH_DEBUG
		printf("Method %i: %x\n",i,method);
#endif
		tmp=authenticators->getPtrValue(method);
		if (tmp&&(!processor)) {
			processor=(proc_auth)tmp;
			chosenmethod=method;
#ifdef WITH_DEBUG
			printf("Chosen method %i\n",method);
#endif
		}
	}
	szResp[0]=0x05;
	szResp[1]=chosenmethod;
	conn->sendData(szResp,2);
	if (chosenmethod!=0xff) {
		return processor(conn);
	} else return FALSE;
}

BOOL Ver5Request(Connection * conn)
{
	unsigned char version;
	unsigned char cmd;
	unsigned char res;
	unsigned char atyp;
	void * tmp;

	if (!conn->getChar(&version)) return FALSE;
	if (version!=5) return FALSE;
	if (!conn->getChar(&cmd)) return FALSE;
	if (!conn->getChar(&res)) return FALSE;
	if (!conn->getChar(&atyp)) return FALSE;
	tmp=conn->getResolver(atyp);
	if (tmp) {
		proc_resolve resolver;
#ifdef WITH_DEBUG
		printf("Applying name resolver %x\n",atyp);
#endif
		resolver=(proc_resolve)tmp;
		resolver(conn);
	} else {
#ifdef WITH_DEBUG
		printf("No resolver\n");
#endif 
		conn->sendResponsev5(0x08);
		return FALSE;
	}
	if (conn->getDestAddressLength()==0) {
#ifdef WITH_DEBUG
		printf("Resolver failed\n");
#endif 
		conn->sendResponsev5(0x04); /* Host unreachable */
		return FALSE;
	}
	if (!conn->IsAllowed(TYPE_DEST)) {
		return FALSE;
	}
	tmp=v5_commands->getPtrValue(cmd);
	if (tmp) {
		BOOL ret;
		proc_command command;
#ifdef WITH_DEBUG
		printf("Applying command\n");
#endif
		command=(proc_command)tmp;
		ret=command(conn);
		return ret;
	} else {
#ifdef WITH_DEBUG
		printf("No command %x\n",cmd);
#endif
		conn->sendResponsev5(0x07);
		return FALSE;
	}
	return FALSE;
}

BOOL cpp_version_process(Connection * conn)
{
	if (!Initialize(conn)) return FALSE;
	return Ver5Request(conn);
}

void ver5_init_module()
{
	authenticators=new DSHashTable();
	v5_commands=new DSHashTable();
	DSComponent auth(authenticators);
	DSComponent com(v5_commands);
	int i;
	i=auth.loadComponentsPrefixed(componentdir "/version5/authenticators", "authenticator_type",
			"auth");
#ifdef WITH_DEBUG
	printf("Loaded %i authenticator components\n",i);
#endif
	i=com.loadComponentsPrefixed(componentdir "/version5/commands",
			"v5command_type", "v5command");
#ifdef WITH_DEBUG
	printf("Loaded %i v5 command components\n",i);
#endif
}

extern "C" {
int libv5_version_type()
{
	ver5_init_module();
	return 5;
}

BOOL libv5_version_process(void * conn)
{
	return cpp_version_process((Connection *)conn);
}
}

